"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GlobalStore = exports.isPrimitive = void 0;
const tslib_1 = require("tslib");
const react_dom_1 = tslib_1.__importDefault(require("react-dom"));
const lodash_1 = require("lodash");
const react_1 = require("react");
const isPrimitive = (value) => lodash_1.isNil(value) || lodash_1.isNumber(value) || lodash_1.isBoolean(value) || lodash_1.isString(value) || typeof value === 'symbol';
exports.isPrimitive = isPrimitive;
/**
* This is a class to create global-store objects
* @template IState
* @param {IState} state - Initial state,
* @template IActions
* @param {IActions} actions - An specific api to restrict the use of the state,
* this will disable the default return of the state-setter of the hook, and instead will return the API
* @param {string} persistStoreAs - A name if you want to persist the state of the store in localstorage
* */
class GlobalStore {
    constructor(state, actions = null, persistStoreAs = null) {
        this.state = state;
        this.actions = actions;
        this.persistStoreAs = persistStoreAs;
        this.subscribers = [];
        this.storedStateItem = undefined;
        this.getPersistStoreValue = () => this.getStoreItem();
        this.getStateCopy = () => Object.freeze(lodash_1.cloneDeep(this.state));
        /**
         * Returns a global hook that will share information across components by subscribing them to a specific store.
         * @return [currentState, GlobalState.IHookResult<IState, IActions, IApi>]
         */
        this.getHook = () => () => {
            const [value, setter] = react_1.useState(this.state);
            const valueWrapper = this.isPersistStore ? this.getPersistStoreValue() : value;
            react_1.useEffect(() => {
                this.subscribers.push(setter);
                return () => {
                    this.subscribers = this.subscribers.filter((hook) => setter !== hook);
                };
            }, []);
            return [
                valueWrapper,
                this.stateOrchestrator,
            ];
        };
        /**
         * This is an access to the subscribers queue and to the current state of a specific store...
         * THIS IS NOT A REACT-HOOK, so you could use it everywhere example other hooks, and services.
         * @return [currentState, GlobalState.IHookResult<IState, IActions, IApi>]
         */
        this.getHookDecoupled = () => {
            const valueWrapper = this.isPersistStore ? () => this.getPersistStoreValue() : () => this.state;
            return [
                valueWrapper,
                this.stateOrchestrator,
            ];
        };
        this._stateOrchestrator = null;
        this.globalSetter = (setter, callback) => {
            const partialState = typeof setter === 'function' ? setter(this.getStateCopy()) : setter;
            let newState = exports.isPrimitive(partialState) || Array.isArray(partialState)
                ? partialState : Object.assign(Object.assign({}, this.state), partialState);
            // avoid perform multiple update batches by accumulating state changes of the same hook
            GlobalStore.batchedUpdates = GlobalStore.batchedUpdates.filter(([, hook, previousState]) => {
                const isSameHook = hook === this;
                if (isSameHook) {
                    // eslint-disable-next-line no-console
                    console.warn('You should try avoid call the same state-setter multiple times at one execution line');
                    newState = exports.isPrimitive(newState) || Array.isArray(partialState)
                        ? newState : Object.assign(Object.assign({}, previousState), newState);
                }
                return !isSameHook;
            });
            this.state = newState;
            // batch store updates
            GlobalStore.batchedUpdates.push([() => this.subscribers.forEach((updateChild) => updateChild(newState)), this, newState]);
            GlobalStore.ExecutePendingBatchesCallbacks.push(callback);
            GlobalStore.ExecutePendingBatches();
        };
        this.globalSetterAsync = (setter) => tslib_1.__awaiter(this, void 0, void 0, function* () { return new Promise((resolve) => this.globalSetter(setter, () => resolve())); });
        this.globalSetterToPersistStoreAsync = (setter) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.globalSetterAsync(setter);
            this.setStoreItem();
        });
        this.getActions = () => {
            const actions = this.actions;
            // Setter is allways async because of the render batch
            const setter = this.isPersistStore ? this.globalSetterToPersistStoreAsync : this.globalSetterAsync;
            return Object.keys(actions).reduce((accumulator, key) => (Object.assign(Object.assign({}, accumulator), { [key]: (...parameres) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                    let promise;
                    const setterWrapper = (value) => {
                        promise = setter(value);
                        return promise;
                    };
                    const result = actions[key](...parameres)(setterWrapper, this.getStateCopy());
                    const resultPromise = Promise.resolve(result) === result ? result : Promise.resolve();
                    yield Promise.all([promise, resultPromise]);
                    return result;
                }) })), {});
        };
    }
    get isPersistStore() {
        return !!this.persistStoreAs;
    }
    formatItemFromStore(obj) {
        const isArray = Array.isArray(obj);
        if (isArray) {
            return obj.map((item) => this.formatItemFromStore(item));
        }
        return Object.keys(obj).filter((key) => !key.includes('_type')).reduce((acumulator, key) => {
            const type = obj[`${key}_type`];
            const unformatedValue = obj[key];
            const isDateType = type === 'date';
            if (isDateType) {
                return Object.assign(Object.assign({}, acumulator), { [key]: new Date(unformatedValue) });
            }
            return Object.assign(Object.assign({}, acumulator), { [key]: exports.isPrimitive(unformatedValue) ? unformatedValue : this.formatItemFromStore(unformatedValue) });
        }, {});
    }
    localStorageGetItem() {
        return localStorage.getItem(this.persistStoreAs);
    }
    getStoreItem() {
        if (this.storedStateItem !== undefined)
            return this.storedStateItem;
        const item = this.localStorageGetItem();
        if (item) {
            const value = JSON.parse(item);
            const primitive = exports.isPrimitive(value);
            const newState = primitive ? value : this.formatItemFromStore(value);
            this.state = primitive || Array.isArray(value) ? newState : Object.assign(Object.assign({}, this.state), newState);
        }
        return this.state;
    }
    formatToStore(obj) {
        const isArray = Array.isArray(obj);
        if (isArray) {
            return obj.map((item) => this.formatToStore(item));
        }
        return Object.keys(obj).reduce((acumulator, key) => {
            const value = obj[key];
            const isDatetime = value instanceof Date;
            return (Object.assign(Object.assign({}, acumulator), { [key]: exports.isPrimitive(value) || isDatetime ? value : this.formatToStore(value), [`${key}_type`]: isDatetime ? 'date' : typeof value }));
        }, {});
    }
    localStorageSetItem(valueToStore) {
        localStorage.setItem(this.persistStoreAs, valueToStore);
    }
    setStoreItem() {
        if (this.storedStateItem === this.state)
            return;
        this.storedStateItem = this.state;
        const valueToStore = exports.isPrimitive(this.state) ? this.state : this.formatToStore(lodash_1.cloneDeep(this.state));
        this.localStorageSetItem(JSON.stringify(valueToStore));
    }
    get stateOrchestrator() {
        if (this._stateOrchestrator)
            return this._stateOrchestrator;
        if (this.actions) {
            this._stateOrchestrator = this.getActions();
        }
        else if (this.persistStoreAs) {
            this._stateOrchestrator = this.globalSetterToPersistStoreAsync;
        }
        else {
            this._stateOrchestrator = this.globalSetterAsync;
        }
        return this._stateOrchestrator;
    }
}
exports.GlobalStore = GlobalStore;
/**
 **  [subscriber-update-callback, hook, newState]
 */
GlobalStore.batchedUpdates = [];
GlobalStore.ExecutePendingBatchesCallbacks = [];
// avoid multiples calls to batchedUpdates
GlobalStore.ExecutePendingBatches = lodash_1.debounce(() => {
    const reactBatchedUpdates = react_dom_1.default.unstable_batchedUpdates || ((mock) => mock());
    reactBatchedUpdates(() => {
        GlobalStore.batchedUpdates.forEach(([execute]) => {
            execute();
        });
        GlobalStore.batchedUpdates = [];
        GlobalStore.ExecutePendingBatchesCallbacks.forEach((callback) => callback());
        GlobalStore.ExecutePendingBatchesCallbacks = [];
    });
}, 0);
exports.default = GlobalStore;
